/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2005, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

RIFF WAVE compensator-compensates a file consisting of a time series
of numbers-each record in the time series is two tab delimited
numbers; the first number is the PCM magnitude for stereo channel one,
the second number is the PCM magnitude for stereo channel two; the
output is a time series, compensated for use with stereo headphones.

Tsheadphone compensates the output of tsriff(1) for headphones, it
takes the time series output of tsriff(1) and compensates it.  For
example:

    tsriff myfile.wav | tsheadphone | tsunriff myfile.wav new.wav

where the file, new.wav, is a RIFF/WAVE file, myfile.wav, compensated
for headphones-the header for the file, new.wav, was copied from the
original, myfile.wav by the tsunriff program, (however, the -H option
to the tsriff(1) program could be used if the header needs to be
manipulated.)

Consider listening to a stereo system with speakers. At low
frequencies, at half wavelengths longer than the distance between the
ears, sound will diffract around the head and there will be no
directional characteristics. So, for wavelengths longer than about a
foot, (corresponding to a frequency of about 1100 Hz.,) there is no
stereo effect. For frequencies below about 1100 Hz., the left and
right channels can be simply added together for each headphone
channel.

Now consider frequencies higher than about 1100 Hz. Those frequencies
have directional characteristics; if the sound from the left and right
speakers are identical, (i.e., monophonic,) the sound would appear to
come from a point source directly in front of the listener-for both
speakers and headphones. If the sound from the left and right speakers
are not identical, then they would appear to come from point sources
to the left and right of a listener with headphones, (which makes the
music appear to come from inside the head.)  Note that it is a vector
addition-identical sound in both channels appears to come from
straight ahead in headphones, while different sounds appear to come
directly from the left or right. To make sound appear to come from in
between those directions, a fraction, the cross feed factor, of one
channel must be added to the other, and vice versa.

A standard configuration for a stereo system is for the speakers to be
located 15 feet apart, and aimed at the listener at an included angle
of 60 degrees, which provides an illusion of spatially distributed
music. For example, if the listener is looking straight ahead, then
the sound level of the right channel would be cos (60) = 0.5 from the
right speaker, and cos (120) = -0.5 from the left speaker. If the
listener's head turns 15 degrees toward the left speaker, the sound
level would raise to cos (45) = 0.707 from the right speaker and cos
(105) = -0.269 from the left speaker. Likewise, sound that is balanced
in a ratio of 0.707 from the right speaker and -0.269 from the left
speaker would appear 15 degrees to the right of the listener, (by
reciprocity arguments.) This is the stereo illusion. Most recorded
music is mixed according to this model.

Looked at from a slightly different perspective, suppose a recording
engineer wants the illusion that a sound is coming from a source that
is theta degrees to the right of straight ahead. Then mixing the sound
level in a ratio of sin(theta) / cos(theta) = tan(theta) between the
left and right channels would produce the desired effect. For example,
if theta was 15 degrees, then sin(15) = 0.259 and cos(15) = 0.966, or
the ratio of right to left channel sound level would be 0.966 / 0.259
= 0.373. (Assuming that the speakers were "spherical radiators.")

For headphones to sound like the same stereo illusion-by reciprocity
arguments-then the sound should appear to come from two sources
separated by +/- 30 degrees, (or 60 and 120 degrees from the axis of
the listener's ears.) Recorded music mixed such that a sound level
ratio of sin (60) / cos (60) = 1.732 between the left and right
speakers would produce sound appearing to come from 30 degrees off
axis. Or, a factor of 1 / 1.732 = 0.577 from one channel should be
added to the other, and vice versa.

So as an approximate model, at frequencies below about 1100 Hz., both
the right and left channels are added together for each stereo
channel. At frequencies above 1100 Hz., each stereo channel has a
cross feed of about a factor of 0.577 = 5 Db. from the other channel
added to it-ignoring sound blocking by the head itself.

Note that the analysis is in good agreement with the empirical (See:
http://headwize.com/tech/elemnts_tech.htm, Technical Paper: The
Elements of Musical Perception by HeadWize for particulars)
literature:

    E. MacPherson, "A Computer Model of Binaural Localiztion of Stereo
    Imaging Measuremnt," JAES, September, 1991

claims about a -3 Db. = 0.71, cross feed below 700 Hz. falling to
about -10.0 Db. = 0.32 cross feed above 700 Hz., with a 300
microsecond delay, (see: http://headwize.com/tech/headrm1_tech.htm,
Technical Paper: The Elements of Muscial Perception, corresponding to
about a quarter of a foot,) in the cross feed between channels. (Which
is about -3Db. at all frequencies from the analysis-the above analysis
did not include blocking by the head, which apparently is about -3Db.)

To produce the single pole low pass filter at 700 Hz., the
http://localhost/www.johncon.com/ndustrix/utilities/tspole.tar.gz
sources from the
http://localhost/www.johncon.com/ndustrix/utilities/tspole.txt
tspole(1) program from the
href="http://localhost/www.johncon.com/ndustrix/">NdustriX site was
used.

The single pole low pass filtering of a time series is implemented
from the following discrete time equation:

    v      = I * k2 + v  * k1
     n + 1             n

where I is the value of the current input sample in the time series, v
are the n'th and n + 1'th value of the output time series, and k1 and
k2 are constants determined from the following equations:

          -2 * p * pi
    k1 = e

and

    k2 = 1 - k1

where p is a constant that determines the frequency of the pole-a
value of unity places the pole at the sample frequency of the time
series.

For a pole frequency of 700 Hz. the value of p is about 0.016, (p =
700 / 44,100,) and pi is 3.1459 ...

The high frequency zero is constructed by multiplying the magnitude of
one channel by 0.32, summing it into the other channel, for all
frequencies; for frequencies below 700 Hz., the magnitude of one
channel is multiplied by 0.71 - 0.32 = 0.39, summing it into the other
channel, (e.g., for frequencies below 700 Hz., the left channel has a
sound level of L + 0.32R + 0.39R = L + 0.71R,) and above 700 Hz., a
factor of 0.32 of one channel is summed into the other, (or the left
channel would be L + 0.32R.) The schematic for the left channel would
be:

            +---+                      +-----+   +---+
    R >-+-->| X |--------------------->| sum |-->| X |--> L'
        |   +---+                      +-----+   +---+
        |     ^                          ^ ^       ^
        |     |                          | |       |
        |     |                          | |       |
        |   0.32                         | |     0.59
        |                                | |
        |   +-----------------+   +---+  | |
        +-->| pole at 700 Hz. |-->| X |--+ |
            +-----------------+   +---+    |
                                    ^      |
                                    |      |
                                    |      |
                                   0.39    |
    L >------------------------------------+

Likewise for the right channel. The gain is adjusted by multiplying
the output by 0.59 to bring the low frequency sound level back near
the original.

Unfortunately, RIFF/WAVE PCM values are restricted to -32768 to 32767,
and since, for frequencies below the pole of 700 Hz., the magnitude
has been increased by a factor of 1.71, the gain is reduced by a
factor of 0.59. This means, for a mono recording, (i.e., both stereo
channels contain the same PCM values,) at frequencies much larger than
700 Hz., the magnitude would be reduced by a factor of 1 + 0.32 / 1.71
= 0.77, or about 2 Db., (voltage or sound pressure, i.e., 20 times log
base 10 the ratio.)

The delay in the cross feed can be altered using the -d option to the
tsheadphone(1) program.  The delay equivalent to 0.25 feet, would be
about 300 microseconds. The delay is implemented with a ring buffer,
each "bucket" representing 1 /44.1 KHz., so 500 microseconds would
require 12.3 buckets, or about 13-this delay is inserted in the
total cross feed, (including below 700 Hz. and above,) for both
channels.

The -p option causes the program to print data, including the minimum,
maximum, and RMS of all PCM values in both channels. Because the
RIFF/WAVE standard only supports PCM data values that are a short
signed int, (two 8 bit bytes,) overflow is regularly possible, which
can be remedied by lowering the value of the -g parameter, and
re-running the program.

$Revision: 0.0 $
$Date: 2005/01/13 13:47:23 $
$Id: tsheadphone.c,v 0.0 2005/01/13 13:47:23 john Exp $
$Log: tsheadphone.c,v $
Revision 0.0  2005/01/13 13:47:23  john
Initial version

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#ifdef __STDC__

#include <float.h>

#endif

#ifndef PI /* make sure PI is defined */

#define PI 3.141592653589793 /* pi to 15 decimal places as per CRC handbook */

#endif

static char rcsid[] = "$Id: tsheadphone.c,v 0.0 2005/01/13 13:47:23 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2005, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define TOKEN_SEPARATORS " \t\n\r\b," /* file record field separators */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Compensate a time series from a time series file from a music CD\n",
    "for headphones\n",
    "Usage: tsheadphone [-b lfcf] [-c hfcf] [-d dly] [-f pole] [-g gain]\n",
    "                   [-p] [-v] [infile.txt [outfile.txt]]\n",
    "    -b lfcf, low frequency cross feed between stereo channels, (0.71)\n",
    "    -c hfcf, high frequency cross feed between stereo channels, (0.32)\n",
    "    -d dly, delay, in seconds, of cross feed between stereo channels, (0.0003)\n",
    "    -f pole, pole frequency, (700)\n",
    "    -g gain, volume gain, (1.0)\n",
    "    -p, print data about PCM values\n",
    "    -v, print the program's version information\n"
    "    infile.txt, input filename\n",
    "    outfile.wav, output filename\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening file\n",
    "Error closing file\n",
    "Error writing to file\n",
    "Error allocating memory\n",
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define ECLOSE 3
#define EWRITE 4
#define EMEMORY 5

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static int strtoken (char *string, char *parse_array, char **parse, char *delim);

#else

static void print_message (); /* print any error messages */
static int strtoken ();

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* input buffer */
         outbuffer[BUFLEN], /* output buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         token_separators[] = TOKEN_SEPARATORS;

    int c, /* command line switch */
        d = (int) 13, /* delay, in 1 / 44.1 KHz., of the cross feed between stereo channels, 295 microseconds is the default */
        delay_ctr = (int) 0, /* delay counter */
        fields, /* number of fields in a record */
        channel1 = (int) 0, /* PCM value for stereo channel 1 */
        channel2 = (int) 0, /* PCM value for stereo channel 2 */
        out1 = (int) 0, /* output of channel 1 */
        out2 = (int) 0, /* output of channel 2 */
        maximum = 0, /* maximum value of any PCM value */
        minimum = 0, /* minimum value of any PCM value */
        count = 0, /* PCM values counter */
        P = 0, /* print data flag, 0 = no, 1 = yes */
        retval = NOERROR; /* return value, assume no error */

    short signed int *delay1 = (short signed int *) 0, /* reference to array of past values of channel 1 */
                     *delay2 = (short signed int *) 0; /* reference to array of past values of channel 2 */

    double f = (double) 0.0158730159, /* pole frequency factor */
           C = (double) 0.316227766, /* high frequency cross feed, default is -10 Db. */
           B = (double) 0.707945784, /* low frequency cross feed factor, default is -3 Db. */
           b = (double) 0.707945784, /* low frequency cross feed, default is -3 Db. */
           k1, /* coefficient k1 in the recursive pole filters */
           k2, /* coefficient k2 in the recursive pole filters */
           g = (double) 0.494028785, /* volume gain, default is 1 / (1 + 0.707945784 + 0.316227766) */
           G = (double) 1.0, /* volume gain, default is 1.0 */
           lastvalue1 = (double) 0.0, /* channel 1 last value of pole */
           lastvalue2 = (double) 0.0, /* channel 2 last value of pole */
           sumsquared = (double) 0.0; /* running value of cumulative sum of squares */

    FILE *infile = (FILE *) stdin, /* input file, default to STDIN */
         *outfile = (FILE *) stdout; /* output file, default to STDOUT */

    while ((c = getopt (argc, argv, "b:c:d:f:g:pv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'b': /* request to change low frequency cross feed? */

                b = atof (optarg); /* yes, save the low frequency cross feed */

                if (b < (double) 0.0 || b > (double) 1.0) /* check for possible values, 0 to 100% */
                {
                    optind = argc; /* force argument error */
                    retval = EARGS; /* assume not enough arguments */
                }

                break;

            case 'c': /* request to change high frequency cross feed? */

                C = atof (optarg); /* yes, save the high frequency cross feed */

                if (C < (double) 0.0 || C > (double) 1.0) /* check for possible values, 0 to 100% */
                {
                    optind = argc; /* force argument error */
                    retval = EARGS; /* assume not enough arguments */
                }

                break;

            case 'd': /* request for delay, in 1 / 44.1 KHz., of the cross feed between stereo channels? */

                d = (int) floor ((atof (optarg) * (double) 44100.0)); /* yes, save the delay, in 1 / 44.1 KHz., of the cross feed between stereo channels */

                if (d < 0 || d > 44100) /* check for possible values, 1 second is the maximum value */
                {
                    optind = argc; /* force argument error */
                    retval = EARGS; /* assume not enough arguments */
                }

                break;

            case 'f': /* request to change pole increase/decrease factor? */

                f = atof (optarg) / (double) 44100.0; /* yes, save the pole frequency factor */

                if (f < (double) 0.000022675737 || f > (double) 0.5) /* check for possible values, 0 to 22050 Hz. */
                {
                    optind = argc; /* force argument error */
                    retval = EARGS; /* assume not enough arguments */
                }

                break;

            case 'g': /* request to change volume gain? */

                G = atof (optarg); /* yes, save the volume gain */

                if (G < (double) 0.0 || G > (double) 10.0) /* check for possible values, 0 to a factor of 10, which is arbitrary */
                {
                    optind = argc; /* force argument error */
                    retval = EARGS; /* assume not enough arguments */
                }

                break;

            case 'p': /* request for print data, 0 = no, 1 = yes? */

                P = 1; /* yes, set the print data flag, 0 = no, 1 = yes */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if ((B = b - C) < (double) 0.0) /* subtract the value of the high frequency cross feed from the low frequency cross feed */
    {
        optind = argc; /* force argument error */
        retval = EARGS; /* assume not enough arguments */
    }

    g =  G * ((double) 1.0 / ((double) 1.0 + B + C)); /* set the volume gain */

    if (g < (double) 0.0 || g > (double) 10.0)
    {
        optind = argc; /* force argument error */
        retval = EARGS; /* assume not enough arguments */
    }

    if (retval == NOERROR) /* enough arguments? */
    {
        k1 = exp (- ((double) 2.0 * (double) PI * f)); /* calculate the coefficient k1 in the recursive filter */
        k2 = (double) 1.0 - k1; /* calculate the coefficient k2 in the recursive filter */

        if (d != (int) 0) /* delay, in 1 / 44.1 KHz., of the cross feed between stereo channels specified */
        {

            if ((delay1 = (short signed int *) malloc ((size_t) d * sizeof (short signed int))) == (short signed int *) 0) /* allocate the array of past pre-compensated values of channel 1 */
            {
                retval = EMEMORY; /* error allocating memory */
            }

            if (retval == NOERROR) /* any errors? */
            {

                if ((delay2 = (short signed int *) malloc ((size_t) d * sizeof (short signed int))) == (short signed int *) 0) /* allocate the array of past pre-compensated values of channel 2 */
                {
                    retval = EMEMORY; /* error allocating memory */
                }

                if (retval == NOERROR) /* any errors? */
                {

                    for (delay_ctr = 0; delay_ctr < (int) d; delay_ctr ++) /* initialize the array with zero values */
                    {
                        delay1[delay_ctr] = delay2[delay_ctr] = (short signed int) 0.0;
                    }

                    delay_ctr = 0; /* and initialize the delay counter */
                }

            }

        }

        if (retval == NOERROR) /* enough arguments? */
        {
            retval = EOPEN; /* assume error opening file */

            if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* open the input file */
            {
                optind ++; /* next argument, if any */

                if ((outfile = argc <= optind ? stdout : fopen (argv[optind], "w")) != (FILE *) 0) /* open the output file */
                {
                    retval = NOERROR; /* assume no errors */

                    while (fgets (buffer, BUFLEN, infile) != (char *) 0) /* read the records from the input file */
                    {

                        if ((fields = strtoken (buffer, parsebuffer, token, token_separators)) != 0) /* parse the record into fields, skip the record if there are no fields */
                        {

                            if (token[0][0] != '#') /* if the first character of the first field is a '#' character, skip it */
                            {

                                if (fields == (int) 2) /* skip the record if it does not have exactly two fields */
                                {
                                    channel1 = (unsigned short int) atoi (token[0]); /* save the PCM value of stereo channel 1 */
                                    channel2 = (unsigned short int) atoi (token[1]); /* save the PCM value of stereo channel 2 */

                                    if (d == (int) 0) /* delay, in 1 / 44.1 KHz., of the cross feed between stereo channels specified, specified? */
                                    {
                                        out1 = (int) floor (((((double) channel1 + (B * lastvalue2) + ((double) channel2 * C)) * g) + (double) 0.5));
                                        lastvalue2 = ((double)((double) channel2) * k2) + (lastvalue2 * k1);

                                        out2 = (int) floor (((((double) channel2 + (B * lastvalue1) + ((double) channel1 * C)) * g) + (double) 0.5));
                                        lastvalue1 = ((double) ((double) channel1) * k2) + (lastvalue1 * k1);
                                    }

                                    else
                                    {
                                        out1 = (int) floor (((((double) channel1 + (B * lastvalue2) + ((double) delay2[delay_ctr] * C)) * g) + (double) 0.5));
                                        lastvalue2 = ((double) ((double) delay2[delay_ctr]) * k2) + (lastvalue2 * k1);

                                        out2 = (int) floor (((((double) channel2 + (B * lastvalue1) + ((double) delay1[delay_ctr] * C)) * g) + (double) 0.5));
                                        lastvalue1 = ((double) ((double) delay1[delay_ctr]) * k2) + (lastvalue1 * k1);

                                        delay1[delay_ctr] = (short int) channel1; /* save the PCM value of stereo channel 1 */
                                        delay2[delay_ctr] = (short int) channel2; /* save the PCM value of stereo channel 2 */
                                        delay_ctr ++; /* increment the delay counter */

                                        if (delay_ctr >= d) /* but don't increment past the end of the array of past values, (e.g., they are circular ring buffers */
                                        {
                                            delay_ctr = 0;
                                        }

                                    }

                                    /*

                                    PCM values are -32768 to 32767

                                    */

                                    if (out1 > (int) 32767.0)
                                    {
                                        (void) fprintf (stderr, "Maximum PCM value exceeded, %d, stereo channel 1\n", out1);
                                        out1 = (int) 32767.0;
                                    }

                                    if (out1 < (int) -32768.0)
                                    {
                                        (void) fprintf (stderr, "Minimum PCM value exceeded, %d, stereo channel 1\n", out1);
                                        out1 = (int) -32768.0;
                                    }

                                    if (out2 > (int) 32767.0)
                                    {
                                        (void) fprintf (stderr, "Maximum PCM value exceeded, %d, stereo channel 2\n", out2);
                                        out2 = (int) 32767.0;
                                    }

                                    if (out2 < (int) -32768.0)
                                    {
                                        (void) fprintf (stderr, "Minimum PCM value exceeded, %d, stereo channel 2\n", out2);
                                        out2 = (int) -32768.0;
                                    }

                                    minimum = (int) floor ((double) out1 < (double) minimum ? (double) out1 : minimum); /* if this is the smallest PCM value, save it */
                                    maximum = (int) floor ((double) out1 > (double) maximum ? (double) out1 : maximum); /* if this is the largest PCM value, save it */
                                    sumsquared = sumsquared + (double) (out1 * out1); /* add the square of the PCM value of the to the running value of cumulative sum of squares */
                                    minimum = (int) floor ((double) out2 < (double) minimum ? (double) out2 : minimum); /* if this is the smallest PCM value, save it */
                                    maximum = (int) floor ((double) out2 > (double) maximum ? (double) out2 : maximum); /* if this is the largest PCM value, save it */
                                    sumsquared = sumsquared + (double) (out2 * out2); /* add the square of the PCM value of the to the running value of cumulative sum of squares */
                                    count = count + 2; /* increment PCM values counter for both channels */

                                    (void) sprintf (outbuffer, "%hd\t%hd\n", (short signed int) out1, (short signed int) out2); /* save the record */

                                    if (fputs (outbuffer, outfile) == EOF) /* print the record */
                                    {
                                        retval = EWRITE; /* error writing to file */
                                        break; /* stop */
                                    }

                                }

                            }

                        }

                    }

                }

                if ((outfile != stdout) && (outfile != (FILE *) 0)) /* output file not STDOUT? */
                {

                    if (fclose (outfile) == EOF) /* no, close the output file */
                    {
                        retval = ECLOSE; /* error closing file */
                    }

                }

            }

            if ((infile != stdin) && (infile != (FILE *) 0)) /* input file not STDIN? */
            {

                if (fclose (infile) == EOF) /* no, close the input file */
                {
                    retval = ECLOSE; /* error closing file */
                }

            }

        }

    }

    if (retval == NOERROR) /* any errors? */
    {

        if (P == 1) /* print data flag, 0 = no, 1 = yes, set? */
        {
            (void) fprintf (stderr, "Maximum PCM value: %d\nMinimum PCM value: %d\nRMS of PCM values: %f\nCount of PCM values: %d\n", maximum, minimum, sqrt (sumsquared / (double) count), count / 2); /* yes, print the data */
        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

parse a record based on sequential delimiters

int strtoken (char *string, char *parse_array, char **parse, char *delim);

parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field, return the
number of fields parsed

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string,
*parse_array,
**parse,
*delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}
