/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2005, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

RIFF WAVE translator-translates a RIFF file from a music CD, into a
time series of numbers-each record in the output time series is two
tab delimited numbers; the first number is the PCM magnitude for
stereo channel one, the second number is the PCM magnitude for stereo
channel two.

Programmers note: the program makes some very basic assumptions about
the construction of a RIFF/WAVE CD music file; it is assumed that the
header is 56 Bytes, (the program just skips it,) and the music data
follows the header, (to the end of the file,) in 2 Byte short integers
per PCM sample, interleaved between stereo channel 1 and channel
2. The program uses a standard ANSI "C" union to read 2 Bytes, (as
unsigned characters,) which is then interpreted as a short signed
integer, (e.g, 2 Bytes,) in the Intel/IBM architecture Byte order.

User's note: the file is tab/table delimited between channel 1 and
channel 2 stereo; to get channel 1, use the Unix cut(1) command, "cut
-f1 ..."  and "cut -f2 ..." for channel 2.

$Revision: 0.0 $
$Date: 2005/01/13 13:47:23 $
$Id: tsriff.c,v 0.0 2005/01/13 13:47:23 john Exp $
$Log: tsriff.c,v $
Revision 0.0  2005/01/13 13:47:23  john
Initial version

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

static char rcsid[] = "$Id: tsriff.c,v 0.0 2005/01/13 13:47:23 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2005, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Construct a time series from a RIFF/WAVE file from a music CD\n",
    "Usage: tsriff [-H headerfile] [-p] [-v] [infile.wav [outfile.txt]]\n",
    "    -H headerfile, header file name\n",
    "    -p, print data about PCM values\n",
    "    -v, print the program's version information\n",
    "    infile.wav, input filename\n",
    "    outfile.txt, output filename\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening file\n",
    "Error closing file\n",
    "Error writing to file\n",
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define ECLOSE 3
#define EWRITE 4

union data /* Union for 2 byte RIFF/WAVE data, Intel/IBM byte architecture is assumed */
{
    short signed int d; /* the data from STDIN as a short signed integer */
    unsigned char c[2]; /* the data from STDIN as a 2 character string */
} data_union;

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */

#else

static void print_message (); /* print any error messages */

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         *headername = (char *) 0; /* reference to header file name */

    int double_byte_ctr = (int) 0, /* double byte counter */
        c, /* command line switch */
        channel1 = (int) 0, /* PCM value for stereo channel 1 */
        channel2 = (int) 0, /* PCM value for stereo channel 2 */
        temp, /* temporary value */
        maximum = 0, /* maximum value of any PCM value */
        minimum = 0, /* minimum value of any PCM value */
        count = 0, /* PCM values counter */
        P = 0, /* print data flag, 0 = no, 1 = yes */
        retval = NOERROR; /* return value, assume no error */

    double sumsquared = (double) 0.0; /* running value of cumulative sum of squares */

    FILE *infile = (FILE *) stdin, /* input file, default to STDIN */
         *outfile = (FILE *) stdout, /* output file, default to STDOUT */
         *headerfile = (FILE *) 0; /* reference to header file */

    while ((c = getopt (argc, argv, "H:pv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'H': /* header file name? */

                headername = optarg; /* save the header file name */
                break;

            case 'p': /* request for print data, 0 = no, 1 = yes? */

                P = 1; /* yes, set the print data flag, 0 = no, 1 = yes */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EOPEN; /* assume error opening file */

        if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* open the input file */
        {
            optind ++; /* next argument, if any */

            if ((outfile = argc <= optind ? stdout : fopen (argv[optind], "w")) != (FILE *) 0) /* open the output file */
            {

                if (headername != (char *) 0) /* make a header file? */
                {

                    if ((headerfile = fopen (headername, "wb")) != (FILE *) 0) /* yes, open the header file */
                    {
                        retval = NOERROR; /* assume no error */
                    }

                }

                else
                {
                    retval = NOERROR; /* assume no error */
                }

                if (retval == NOERROR) /* assume no error */
                {

                    while (fread (&data_union.c, sizeof (char), (size_t) 2, infile) == (size_t) 2 * sizeof (char)) /* read 2 bytes from the file */
                    {

                        if (++ double_byte_ctr > 28) /* skip over RIFF WAVE header of 56 bytes */
                        {
                            temp = data_union.d; /* a PCM value, save it */
                            minimum = temp < minimum ? temp : minimum; /* if this is the smallest PCM value, save it */
                            maximum = temp > maximum ? temp : maximum; /* if this is the largest PCM value, save it */
                            sumsquared = sumsquared + (double) (temp * temp); /* add the square of the PCM value of the to the running value of cumulative sum of squares */
                            count ++; /* increment PCM values counter */

                            if (double_byte_ctr % (int) 2 > (int) 0) /* first audio channel */
                            {

                                if (temp < -32768) /* right values for a PCM value? */
                                {
                                    (void) fprintf (stderr, "Minimum PCM value exceeded, %d, stereo channel 1\n", temp);
                                    temp = data_union.d = (int) -32768; /* fix it */
                                }

                                if (temp > 32767) /* right values for a PCM value? */
                                {
                                    (void) fprintf (stderr, "Maximum PCM value exceeded, %d, stereo channel 1\n", temp);
                                    temp = data_union.d = (int) 32767; /* fix it */
                                }

                                channel1 = temp; /* save the PCM value for stereo channel 1 */
                            }

                            else
                            {

                                if (temp < -32768) /* right values for a PCM value? */
                                {
                                    (void) fprintf (stderr, "Minimum PCM value exceeded, %d, stereo channel 2\n", temp);
                                    temp = data_union.d = (int) -32768; /* fix it */
                                }

                                if (temp > 32767) /* right values for a PCM value? */
                                {
                                    (void) fprintf (stderr, "Maximum PCM value exceeded, %d, stereo channel 2\n", temp);
                                    temp = data_union.d = (int) 32767; /* fix it */
                                }

                                channel2 = temp; /* save the PCM value for stereo channel 2 */
                                (void) sprintf (buffer, "%hd\t%hd\n", (short signed int) channel1, (short signed int) channel2); /* save the record */

                                if (fputs (buffer, outfile) == EOF) /* print the record */
                                {
                                    retval = EWRITE; /* error writing to file */
                                    break; /* stop */
                                }

                            }

                        }

                        else
                        {

                            if (headerfile != (FILE *) 0) /* making a header file? */
                            {
                                retval = EWRITE; /* assume error writing to file */

                                if (fputc ((int) data_union.c[0], headerfile) != EOF) /* write the first byte to the header file */
                                {

                                    if (fputc ((int) data_union.c[1], headerfile) != EOF) /* write the second byte to the header file */
                                    {
                                        retval = NOERROR; /* assume no errors */
                                    }

                                }

                                if (retval != NOERROR) /* write error? */
                                {
                                    break; /* yes, stop */
                                }

                            }

                        }

                    }

                }

                if (headerfile != (FILE *) 0) /* header file opened? */
                {

                    if (fclose (headerfile) == EOF) /* no, close the header file */
                    {
                        retval = ECLOSE; /* error closing file */
                    }

                }

            }

            if ((outfile != stdout) && (outfile != (FILE *) 0)) /* output file not STDOUT? */
            {

                if (fclose (outfile) == EOF) /* no, close the output file */
                {
                    retval = ECLOSE; /* error closing file */
                }

            }

        }

        if ((infile != stdin) && (infile != (FILE *) 0)) /* input file not STDIN? */
        {

            if (fclose (infile) == EOF) /* no, close the input file */
            {
                retval = ECLOSE; /* error closing file */
            }

        }

    }

    if (retval == NOERROR) /* any errors? */
    {

        if (P == 1) /* print data flag, 0 = no, 1 = yes, set? */
        {
            (void) fprintf (stderr, "Maximum PCM value: %d\nMinimum PCM value: %d\nRMS of PCM values: %f\nCount of PCM values: %d\n", maximum, minimum, sqrt (sumsquared / (double) count), count / 2); /* yes, print the data */
        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}
