/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2005, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

RIFF WAVE translator-translates a file consisting of a time series of
numbers-each record in the time series is two tab delimited numbers;
the first number is the PCM magnitude for stereo channel one, the
second number is the PCM magnitude for stereo channel two; the output
is a RIFF file for a music CD track.

Tsunriff does the exact opposite of tsriff(1), it takes the time
series output of tsriff(1) and makes a RIFF/WAVE file out of it. For
example:

    tsriff myfile.wav | tsunriff myfile.wav new.wav

where the files, myfile.wav and new.wav, should be identical-the
header for the file, new.wav, was copied from the original, myfile.wav
by the tsunriff program, (however, the -H option to the tsriff(1)
program could be used if the header needs to be manipulated.)

Other programs may be placed in the "pipe" for digital filtering,
etc. The output file size of the tsriff(1) program is very large,
(sometimes, hundreds of MB,) and a Unix "pipe" is an attractive
alternative methodology for dealing with large files.

There is one mandatory argument, the name of the header file. The
header file contains the first 56 Bytes of the RIFF/WAVE music file,
and could have been made with the tsriff(1) program, or copied from
the original RIFF/WAVE CD track.

Programmers note: the program makes some very basic assumptions about
the construction of a RIFF/WAVE CD music file; it is assumed that the
header is 56 Bytes, the program just copies it, and the music data
follows the header, (to the end of the file,) in 2 Byte short integers
per PCM sample, interleaved between stereo channel 1 and channel
2. The program uses a standard ANSI "C" union to read 2 Bytes, (as
unsigned characters,) which is then interpreted as a short signed
integer, (e.g, 2 Bytes,) in the Intel/IBM architecture Byte order.

$Revision: 0.0 $
$Date: 2005/01/13 13:47:23 $
$Id: tsunriff.c,v 0.0 2005/01/13 13:47:23 john Exp $
$Log: tsunriff.c,v $
Revision 0.0  2005/01/13 13:47:23  john
Initial version

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

static char rcsid[] = "$Id: tsunriff.c,v 0.0 2005/01/13 13:47:23 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2005, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define TOKEN_SEPARATORS " \t\n\r\b," /* file record field separators */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Construct a RIFF/WAVE file for a music CD from a time series\n",
    "Usage: tsunriff [-p] [-v] headerfile [infile.txt [outfile.wav]]\n",
    "    headerfile, header file name\n",
    "    -p, print data about PCM values\n",
    "    -v, print the program's version information\n",
    "    infile.txt, input filename\n",
    "    outfile.wav, output filename\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening file\n",
    "Error closing file\n",
    "Error reading file\n",
    "Error writing to file\n",
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define ECLOSE 3
#define EREAD 4
#define EWRITE 5

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static int strtoken (char *string, char *parse_array, char **parse, char *delim);

#else

static void print_message (); /* print any error messages */
static int strtoken ();

#endif

union data /* Union for 2 byte RIFF/WAVE data, Intel/IBM byte architecture is assumed */
{
    short signed int d; /* the PCM data as a short signed integer */
    unsigned char c[2]; /* the PCM data as a 2 character string */
} data_union1, data_union2;

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         token_separators[] = TOKEN_SEPARATORS;

    int c, /* command line switch */
        i, /* character counter */
        ch, /* input character */
        fields, /* number of fields in a record */
        channel1, /* PCM value for stereo channel 1 */
        channel2, /* PCM value for stereo channel 2 */
        maximum = 0, /* maximum value of any PCM value */
        minimum = 0, /* minimum value of any PCM value */
        count = 0, /* PCM values counter */
        P = 0, /* print data flag, 0 = no, 1 = yes */
        retval = NOERROR; /* return value, assume no error */

    double sumsquared = (double) 0.0; /* running value of cumulative sum of squares */

    FILE *infile = (FILE *) stdin, /* input file, default to STDIN */
         *outfile = (FILE *) stdout, /* output file, default to STDOUT */
         *headerfile = (FILE *) 0; /* reference to header file */

    while ((c = getopt (argc, argv, "pv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'p': /* request for print data, 0 = no, 1 = yes? */

                P = 1; /* yes, set the print data flag, 0 = no, 1 = yes */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EARGS; /* assume not enough arguments */

        if (argc > optind) /* enough arguments? */
        {
            retval = EOPEN; /* assume error opening file */

            if ((headerfile = fopen (argv[optind], "r")) != (FILE *) 0) /* open the header file */
            {
                optind ++; /* next argument, if any */

                if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* open the input file */
                {
                    optind ++; /* next argument, if any */

                    if ((outfile = argc <= optind ? stdout : fopen (argv[optind], "w")) != (FILE *) 0) /* open the output file */
                    {
                        retval = NOERROR; /* assume no errors */

                        for (i = 0; i < 56; i ++) /* for each byte in the RIFF/WAVE header */
                        {

                            if ((ch = fgetc (headerfile)) == EOF) /* get a charcter */
                            {
                                retval = EREAD; /* error reading file */
                                break; /* stop */
                            }

                            if (fputc (ch, outfile) == EOF) /* and write the character to the output file */
                            {
                                retval = EWRITE; /* error writing to file */
                                break; /* stop */
                            }

                        }

                        if (retval == NOERROR) /* any errors? */
                        {

                            while (fgets (buffer, BUFLEN, infile) != (char *) 0) /* read the records from the input file */
                            {

                                if ((fields = strtoken (buffer, parsebuffer, token, token_separators)) != 0) /* parse the record into fields, skip the record if there are no fields */
                                {

                                    if (token[0][0] != '#') /* if the first character of the first field is a '#' character, skip it */
                                    {

                                        if (fields == (int) 2) /* skip the record if it does not have exactly two fields */
                                        {
                                            channel1 = atoi (token[0]); /* save the PCM value of stereo channel 1 */
                                            channel2 = atoi (token[1]); /* save the PCM value of stereo channel 2 */
                                            minimum = channel1 < minimum ? channel1 : minimum; /* if this is the smallest PCM value, save it */
                                            maximum = channel1 > maximum ? channel1 : maximum; /* if this is the largest PCM value, save it */
                                            sumsquared = sumsquared + (double) (channel1 * channel1); /* add the square of the PCM value of the to the running value of cumulative sum of squares */
                                            minimum = channel2 < minimum ? channel2 : minimum; /* if this is the smallest PCM value, save it */
                                            maximum = channel2 > maximum ? channel2 : maximum; /* if this is the largest PCM value, save it */
                                            sumsquared = sumsquared + (double) (channel2 * channel2); /* add the square of the PCM value of the to the running value of cumulative sum of squares */
                                            count = count + 2; /* increment PCM values counter for both channels */

                                            if (channel1 < -32768) /* right values for a PCM value? */
                                            {
                                                (void) fprintf (stderr, "Minimum PCM value exceeded, %d, stereo channel 1\n", channel1);
                                                channel1 = (int) -32768; /* fix it */
                                            }

                                            if (channel1 > 32767) /* right values for a PCM value? */
                                            {
                                                (void) fprintf (stderr, "Maximum PCM value exceeded, %d, stereo channel 1\n", channel1);
                                                channel1 = (int) 32767; /* fix it */
                                            }

                                            if (channel2 < -32768) /* right values for a PCM value? */
                                            {
                                                (void) fprintf (stderr, "Minimum PCM value exceeded, %d, stereo channel 2\n", channel2);
                                                channel2 = (int) -32768; /* fix it */
                                            }

                                            if (channel2 > 32767) /* right values for a PCM value? */
                                            {
                                                (void) fprintf (stderr, "Maximum PCM value exceeded, %d, stereo channel 2\n", channel2);
                                                channel2 = (int) 32767; /* fix it */
                                            }

                                            data_union1.d = (short signed int) channel1; /* save the PCM values of the the stereo channels */
                                            data_union2.d = (short signed int) channel2;
                                            retval = EWRITE; /* assume error writing to file */

                                            if (fputc ((int) data_union1.c[0], outfile) != EOF) /* write the first byte to the output file */
                                            {

                                                if (fputc ((int) data_union1.c[1], outfile) != EOF) /* write the second byte to the output file */
                                                {

                                                    if (fputc ((int) data_union2.c[0], outfile) != EOF) /* write the third byte to the output file */
                                                    {

                                                        if (fputc ((int) data_union2.c[1], outfile) != EOF) /* write the forth byte to the output file */
                                                        {
                                                            retval = NOERROR; /* assume no errors */
                                                        }

                                                    }

                                                }

                                            }

                                            if (retval != NOERROR) /* write error? */
                                            {
                                                break; /* yes, stop */
                                            }

                                        }

                                    }

                                }

                            }

                        }

                    }

                    if ((outfile != stdout) && (outfile != (FILE *) 0)) /* output file not STDOUT? */
                    {

                        if (fclose (outfile) == EOF) /* no, close the output file */
                        {
                            retval = ECLOSE; /* error closing file */
                        }

                    }

                }

                if ((infile != stdin) && (infile != (FILE *) 0)) /* input file not STDIN? */
                {

                    if (fclose (infile) == EOF) /* no, close the input file */
                    {
                        retval = ECLOSE; /* error closing file */
                    }

                }

            }

            if (headerfile != (FILE *) 0) /* header file opened? */
            {

                if (fclose (headerfile) == EOF) /* no, close the header file */
                {
                    retval = ECLOSE; /* error closing file */
                }

            }

        }

    }

    if (retval == NOERROR) /* any errors? */
    {

        if (P == 1) /* print data flag, 0 = no, 1 = yes, set? */
        {
            (void) fprintf (stderr, "Maximum PCM value: %d\nMinimum PCM value: %d\nRMS of PCM values: %f\nCount of PCM values: %d\n", maximum, minimum, sqrt (sumsquared / (double) count), count / 2); /* yes, print the data */
        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

parse a record based on sequential delimiters

int strtoken (char *string, char *parse_array, char **parse, char *delim);

parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field, return the
number of fields parsed

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string,
*parse_array,
**parse,
*delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}
